#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   05-Sep-2006

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import time
import math
try:
	import numpy as np
except ImportError:
	np = None
try:
	import matplotlib.pyplot as pyplot
except ImportError:
	pyplot = None

import Data
import Gnuplot
import Evaluation
try:
	import Matplotlib
except ImportError:
	Matplotlib = None

#from Plot import *
from Plot import Plot, PALETTE

#===============================================================================
# ----------------------------------------------------------------------
# Base class to start plotting (Launches gnuplot)
# ----------------------------------------------------------------------
def start():
	if Matplotlib is None: return None
	engine = Matplotlib.Matplotlib()
	while 20:	# wait a while to get version
		try:
			ver = engine.version()
		except:
			return None
		if ver is not None: break

	# Initialize mouse and functions
	#engine("bin(x,m,s)=floor((x-m)/s)*s+m")
	#engine("logbin(x,m,s)=10**(floor((log10(x)-m)/s)*s+m)")
	#engine("rnd(x)=abs(int((x*1103515245+12345)/65536)%32768)")
	return engine

#===============================================================================
# MatPlotlib Plot class
#===============================================================================
class MPPlot(Plot):
	# ----------------------------------------------------------------------
	def __call__(self, s):
		self.write(s)
		self.plot.write(s)

	# ----------------------------------------------------------------------
	# Save plot to file
	# ----------------------------------------------------------------------
	def save(self, ext):
		pass

	# ----------------------------------------------------------------------
	# Send plot on printer
	# ----------------------------------------------------------------------
	def hardcopy(self, printer):
		pass

	# ----------------------------------------------------------------------
	def show(self):
		super().show()
		self.engine.show(False)

	# ----------------------------------------------------------------------
	# @return plot options for prefix: title, x, y, ...
	# ----------------------------------------------------------------------
	def _plotOptions(self, prefix):
		# Find the corresponding one in Matplotlib
		font  = self.plot("%sfont"%(prefix))
		#size  = self.plot("%ssize"%(prefix))
		#color = self.engine.color(self.plot("%scolor"%(prefix)))
		font = {
			# "fontname" : font,
			# "fontsize" : size,
			# "color"    : color
		}
		return font

	# ----------------------------------------------------------------------
	# Set color band
	# ----------------------------------------------------------------------
	def plotColorBand(self, pm3dopts=None):
		# Color band
		try: cbmin = str(float(self.plot("cbmin")))
		except: cbmin = ""

		try: cbmax = str(float(self.plot("cbmax")))
		except: cbmax = ""

		try: cpd = float(self.plot("cbcpd"))
		except: cpd = 3.0

		cblog = int(self.plot("cblog"))

		# If logscale... (not for balance)
		if cbmin!="" or cbmax!="":
			if cblog and int(self.plot("cbround")) and cbmin!="" and cpd>=1.0:
				# round limits that colors boundaries falls on decades
				try:
					nmin = math.floor(math.log10(float(cbmin))*cpd)/cpd
					cbmin = "%g"%(math.pow(10.0, nmin))
				except:
					cbmin = "*"
			self("set cbrange [%s:%s]" % (cbmin,cbmax))

		# Set the pm3d
		try:
			colors = int(self.plot("cbcolors"))
		except ValueError:
			colors = 30
		self.plotPm3D(colors, self.plot("cbpalette"), pm3dopts)

		if cblog:
			self("set logscale cb")
			#if 4.4 < self.engine.version() < 5.0:
			#	self("set logscale z")	# Necessary for 4.4.4+
							# to treat log(0) as 0 and not as undefined

	# ----------------------------------------------------------------------
	# Get axes range from engine
	# ----------------------------------------------------------------------
#	def getAxesRangeFromGnuplot(self):
#		if not self.engine: return
#		xlow, xhigh = self.engine.getRange("x")
#		ylow, yhigh = self.engine.getRange("y")
#		try:
#			zlow, zhigh = self.engine.getRange("z")
#		except:
#			pass
#
#		self.resetAxesRange(False)
#		if xlow is not None:  self.x_min.insert(0, xlow)
#		if ylow is not None:  self.y_min.insert(0, ylow)
#		if xhigh is not None: self.x_max.insert(0, xhigh)
#		if yhigh is not None: self.y_max.insert(0, yhigh)
#
#		try:
#			if zlow is not None:  self.z_min.insert(0, zlow)
#			if zhigh is not None: self.z_max.insert(0, zhigh)
#		except:
#			pass
#
#		try:
#			if self.x2_tics.get():
#				self.x2_min.delete(0,END)
#				self.x2_max.delete(0,END)
#				xlow, xhigh = self.engine.getRange("x2")
#
#				if xlow is not None:  self.x2_min.insert(0, xlow)
#				if xhigh is not None: self.x2_max.insert(0, xhigh)
#
#			if self.y2_tics.get():
#				self.y2_min.delete(0,END)
#				self.y2_max.delete(0,END)
#				ylow, yhigh = self.engine.getRange("y2")
#
#				if ylow is not None:  self.y2_min.insert(0, ylow)
#				if yhigh is not None: self.y2_max.insert(0, yhigh)
#		except:
#			pass

	# ----------------------------------------------------------------------
	# Plot Using
	# return using string for line plots depending on parameters requested
	# ----------------------------------------------------------------------
	def _plotUsing(self, style, x, y, ex, ey):
		try:
			specs = Gnuplot.SPEC[Gnuplot.STYLE.index(style)]
		except ValueError:
			specs = 2
		if   specs == 2:
			return "%s:%s" % (x, y)
		elif specs == 3:
			if style in ["xerrorbars", "xerrorlines"]:
				return "%s:%s:%s" % (x, y, ex)
			else:
				return "%s:%s:%s" % (x, y, ey)
		elif specs == 4:
			return "%s:%s:%s:%s" % (x, y, ex, ey)

	# ----------------------------------------------------------------------
	# Plot Style
	# @return string with plot options
	# ----------------------------------------------------------------------
	def _plotStyle(self, det, style=None):
		# axes
		axes = self.plot("axes.%d"%(det))
		if axes != "x1y1":
			cmd = "axes %s " % (axes)
		else:
			cmd = ""

		# with
		if style is None:
			style = self.plot("with.%d"%(det))
			if style in ("histogram","histerror"): style = "steps"
		cmd += "w %s " % (style)

		# smooth
		smooth = self.plot("smooth.%d"%(det))
		if smooth != "":
			cmd += "smooth %s "%(smooth)

		# line style
		#if style not in ["points", "dots"]:
		x = int(self.plot("ls.%d"%(det)))
		if x:
			cmd += "ls %s "%(x)
		else:
			x = self.plot("lt.%d"%(det))
			if x: cmd += "lt %s "%(x)

			x = self.plot("lw.%d"%(det))
			if x: cmd += "lw %s "%(x)

			if self.engine.version() >= 4.2:
				x = self.plot("lc.%d"%(det))
				if x: cmd += "lc rgb '%s' "%(x)

			# point style
			if style not in ["lines", "steps", "impulses", "fsteps", "dots",
					"filledcurves", "histeps", "boxes",
					"boxerrorbars", "boxxyerrorbars"]:
				cmd += "pt %s ps %s " \
					% (self.plot("pt.%d"%(det)),
					   self.plot("ps.%d"%(det)))

		#if style in ["filledcurves", "boxes", \
		#		"boxerrorbars", "boxxyerrorbars"]:
		#	cmd += "fs %s " % (self.plot("fs.%d"%(det),"1"))
		return cmd

	# ----------------------------------------------------------------------
	# Perform plot
	# ----------------------------------------------------------------------
	def _plot(self, cmd):
		if self.engine.version() == 0: return None
		self.plotCommands()
		if self.plotappend != "":
			if cmd:
				cmd += ",%s"%(self.plotappend)
			else:
				cmd = self.plotappend
		if cmd: self("plot %s"%(cmd))
		return self._plotEnd()

	# ----------------------------------------------------------------------
	# Perform splot
	# ----------------------------------------------------------------------
	def _splot(self, cmd):
		if self.engine.version() == 0: return None
		self.plotCommands()
		if self.plotappend != "":
			if cmd:
				cmd += ",%s"%(self.plotappend)
			else:
				cmd = self.plotappend
		if cmd: self("splot %s"%(cmd))
		return self._plotEnd()

	# ----------------------------------------------------------------------
	# Base class to start plotting (Launches engine)
	# ----------------------------------------------------------------------
	def _plotEnd(self):
		self.plotCommands(False)
		return self.checkOutput()

	# ----------------------------------------------------------------------
	def checkOutput(self):
		self.engine.writeline('print "end"')
		# wait until we get the line end
		line  = "\n"
		END   = "end\n"
		err   = None
		tstart = time.time()
		while line!=END and time.time()-tstart < 10.0:	# 10s
			while self.engine.check(1.0):
				line = self.engine.readline()
				if line not in ("\n", END):
					if "Warning" in line:
						self.log(f">w> {line}")
					else:
						self.log(f">e> {line}")
					if err is None:
						err = "Matplot error or warning"
		return err

	# ----------------------------------------------------------------------
	# Enable pm3d mode for gnuplot
	# ----------------------------------------------------------------------
	def plotPm3D(self, colors=None, palette="FLUKA", pm3dopts=None):
		self("set colorbox vertical")
		if pm3dopts is None:
			self("set pm3d map explicit corners2color c1")
		else:
			self("set pm3d %s"%(pm3dopts))

		if not isinstance(palette, str):
			palette.gnuplot_setPalette(self.engine)
			return
		elif palette == PALETTE[0]:
			self('set palette defined ('
				' 1. 1.0 1.0 1.0,  2. 0.9 0.6 0.9,  3. 1.0 0.4 1.0,'
				' 4. 0.9 0.0 1.0,  5. 0.7 0.0 1.0,  6. 0.5 0.0 0.8,'
				' 7. 0.0 0.0 0.8,  8. 0.0 0.0 1.0,  9. 0.0 0.6 1.0,'
				'10. 0.0 0.8 1.0, 11. 0.0 0.7 0.5, 12. 0.0 0.9 0.2,'
				'13. 0.5 1.0 0.0, 14. 0.8 1.0 0.0, 15. 1.0 1.0 0.0,'
				'16. 1.0 0.8 0.0, 17. 1.0 0.5 0.0, 18. 1.0 0.0 0.0,'
				'19. 0.8 0.0 0.0, 20. 0.6 0.0 0.0, 21. 0.0 0.0 0.0 )' )
		elif palette == PALETTE[1]:
			self('set palette defined ( 1 "grey",'
				'2 "magenta", 3 "blue",4 "green",'
				'5 "yellow", 6 "red", 7 "black")')
		elif palette == PALETTE[2]:
			self('set palette')
		elif palette == "rainbow":
			self('set palette model HSV defined ( 0 0 1 1, 1 1 1 1 )')
		elif palette == "rgb":
			self('set palette defined ( 0 0 0 0, 0.25 0 0 1, 0.50 0 1 0, 0.75 1 0 0, 1 1 1 1 )')
		elif palette == "red-yellow":
			self("set palette model XYZ functions gray**0.35, gray**0.5, gray**0.8")
		elif palette == "geometry":
			self('set palette defined ( 1 "grey",'
				'2 "magenta", 3 "blue",4 "green",'
				'5 "yellow", 6 "red", 7 "white")')
		elif palette == "field":
			# Fluka colors where COLOR = (1+COLOR)/2
			self('set palette defined ('
				' 1.  1.00  1.00  1.00,  2.  0.90  0.90  0.90,'
				' 3.  0.95  0.80  0.95,  4.  1.00  0.70  1.00,'
				' 5.  0.95  0.50  1.00,  6.  0.85  0.50  1.00,'
				' 7.  0.75  0.50  0.90,  8.  0.50  0.50  0.90,'
				' 9.  0.50  0.50  1.00, 10.  0.50  0.80  1.00,'
				'11.  0.50  0.90  1.00, 12.  0.50  0.85  0.75,'
				'13.  0.50  0.95  0.60, 14.  0.75  1.00  0.50,'
				'15.  0.90  1.00  0.50, 16.  1.00  1.00  0.50,'
				'17.  1.00  0.90  0.50, 18.  1.00  0.75  0.50,'
				'19.  1.00  0.50  0.50, 20.  0.90  0.50  0.50,'
				'21.  0.80  0.50  0.50, 22.  0.70  0.50  0.50,'
				'23.  0.50  0.50  0.50 )' )
		elif palette == "Matlab":
			self('set palette defined ('
				'0  0.0 0.0 0.5,'
				'1  0.0 0.0 1.0,'
				'2  0.0 0.5 1.0,'
				'3  0.0 1.0 1.0,'
				'4  0.5 1.0 0.5,'
				'5  1.0 1.0 0.0,'
				'6  1.0 0.5 0.0,'
				'7  1.0 0.0 0.0,'
				'8  0.5 0.0 0.0 )')
		else:
			self('set palette %s'%(palette))

		if colors:
			# Stupid bug fix of gnuplot
			# doesn't work when printing!!!!
			if self.engine.version() < 4.2: colors += 1
			self('set palette maxcolors %d' % (colors))

	# ----------------------------------------------------------------------
	# Set normalization factor
	# ----------------------------------------------------------------------
	def plotNorm(self, norm, column, axis="y"):
		# Normalization and default using
		if norm=="":
			return str(column)
		try:
			int(column)
			column="($%s)"%(column)
		except:
			pass

		try:
			norm = float(eval(norm, Evaluation._globalDict))
			if isinstance(column,int):
				return "($%d*%g)" % (column, norm)
			else:
				return "(%s*%g)" % (column, norm)
		except:
			func=axis+"norm"
			if norm.find("$")>=0:
				return "(%s)"%(norm)
			else:
				self("%s(x)=%s"%(func,norm))
				if isinstance(column,int):
					return "(%s($%d))"%(func,column)
				else:
					return "(%s(%s))"%(func,column)

	# ----------------------------------------------------------------------
	# Plot commands
	# ----------------------------------------------------------------------
	def plotCommands(self, prefix=True):
		first   = True
		pline   = ""
		self.plotappend = ""
		for line in self.plot("commands").splitlines():
			line = line.strip()
			if pline != "":
				line = pline + line
				pline = ""
			if line!="" and line[-1] == "\\":
				pline = line[:-1]
				continue

			words = line.split()
			if not words: continue
			# stop on the first plot command
			if words[0] in ["plot", "replot", "splot"]:
				if words[0] == "plot":
					self.plotappend = line[5:]
				if prefix: break
				if words[0] != "plot":
					self(line+"\n")
				first = False
			elif prefix == first:
				self(line+"\n")

	# ----------------------------------------------------------------------
	# Generate the plot
	# ----------------------------------------------------------------------
	def _geometryPlot(self):
		pass

	# ----------------------------------------------------------------------
	# Generate the plot
	# ----------------------------------------------------------------------
	def _usr1DPlot(self):
		# --- Loop over detectors to build plot string ---
		for i in range(int(self.plot("ndetectors"))):
			if int(self.plot("show.%d"%(i))) != 1:
				continue	# Skip graph
			filename = self.plot("file.%d"%(i))
			if filename == "": continue
			det = int(self.plot("det.%d"%(i)))

			# block to display
			try:
				block = int(self.plot("block.%d"%(i),0))
				if block > 0:
					block -= 1
					det += " every :::%d::%d "%(block,block)
			except:
				pass

			# load data
			try:
				name, xlow, xhigh, y, yerr = Data.tabLis(filename, det, block)
			except:
				# error cannot open file...
				continue

			# ===================================
			# block to display
			# try:
			#	block = int(self.plot("block.%d"%(i),0))
			#	if block > 0:
			#		block -= 1
			#		det += " every :::%d::%d "%(block,block)
			# except:
			#	pass
			# cmd += "'%s' %s " % (filename, det)
			# ===================================

			xlog = int(self.plot("xlog"))
			style = self.plot("with.%d"%(i))
			addErrors = style == "histerror"  # Additional errors

			# Default using (middle of bin)
			if xlog:
				X = np.sqrt(xlow * xhigh)
			else:
				X = (xlow + xhigh) / 2.0

			if style in ("histogram", "histerror"):
				kwargs = {'drawstyle': 'steps-mid'}
				X      = xlow
				usingex= ""
			else:
				kwargs = {}

			yval = int(self.plot("y.%d"%(i)))
			if yval == 1:	# <X>*Y
				Y  = X * y
				EY = X * y * yerr / 100.0
			elif yval == 2:  # DX*Y
				Y  = (xhigh - xlow) * y
				EY = (xhigh - xlow) * y * yerr / 100.0

			else:  # Leave default Y
				Y  = y
				EY = y * yerr / 100.0

			pyplot.plot(X, Y, **kwargs)
			continue

			# Normalization of x-axis
			xnorm = self.plot("xnorm.%d"%(i))
			if xnorm:
				usingx  = self.plotNorm(xnorm, usingx,  "x%d"%(i))
				if usingex:
					usingex = self.plotNorm(xnorm, usingex, "ex%d"%(i))
				if addErrors:
					usingxe = self.plotNorm(xnorm, usingxe, "xe%d"%(i))

			# Normalization and default using
			norm = self.plot("norm.%d"%(i))
			if norm:
				usingy  = self.plotNorm(norm, usingy,  "y%d"%(i))
				usingey = self.plotNorm(norm, usingey, "ey%d"%(i))

			cmd += "us %s "% \
				(self._plotUsing(style, usingx,
					usingy, usingex, usingey))

			# Plotting style
			cmd += self._plotStyle(i, style)

			# Label
			title = self.plot("name.%d"%(i))
			key   = int(self.plot("key.%d"%(i),1))
			if title.strip()=="" or key==0:
				cmd += "notitle"
			else:
				cmd += "t '%s'" % (title)

			if addErrors:
				cmd += ",'' %s us %s "% \
					(det, self._plotUsing("errorbars", usingxe,
						usingy, "", usingey))
				cmd += self._plotStyle(i, "errorbars")
				cmd += "notitle"

			# return self._plot(cmd)

	# ----------------------------------------------------------------------
	# USR-2D plot
	# ----------------------------------------------------------------------
	def _usr2DPlot(self):
		pass

	# ----------------------------------------------------------------------
	# Usrbin plot
	# ----------------------------------------------------------------------
	def _usrbinPlot(self):
		pass

	# ----------------------------------------------------------------------
	# RESNUCLEi plot
	# ----------------------------------------------------------------------
	def _resnucleiPlot(self):
		pass

	# ----------------------------------------------------------------------
	def _userdumpPlot(self):
		pass
