#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   07-Dec-2007

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import log
import csg
import bmath
from deck import Card

_DENSITY_LIMIT = 0.1	# Ligther than that is ignored
#_FAR_AWAY      = 1E10

#-------------------------------------------------------------------------------
# Create rotation matrix using the X,Y,Z vectors
#-------------------------------------------------------------------------------
def write_matrix(fout, X, Y, Z, T):
	M = bmath.Matrix()
	M.make(X,Y,Z)
	M.transpose()
	for i in range(3): M[i][3] = T[i]
	fout.write("\tmultmatrix(m=[\n")
	for r in range(4):
		if r==3: comma=""
		else: comma=","
		fout.write(f"\t\t\t{M[r]}{comma}\n")
	fout.write("\t\t])\n")

#-------------------------------------------------------------------------------
# Create rotation matrix using the Z vector
#-------------------------------------------------------------------------------
def write_matrixZ(fout, Z, T):
	X = Z.orthogonal()
	Y = Z ^ X
	write_matrix(fout,X,Y,Z,T)

##-------------------------------------------------------------------------------
def writePlane0(fout):
	fout.write("\t\tpolyhedron( [[0,INF,0], [INF,-INF/2,0], [-INF,-INF/2,0], [0,0,-INF]],\n")
	fout.write("\t\t\t[[0,1,2], [0,3,1], [1,3,2], [2,3,0]] );")

#-------------------------------------------------------------------------------
# Return a valid povray name
#-------------------------------------------------------------------------------
#def name(s):
#	# First convert all invalid characters to underscore (_)
#	# Minimize the occurrences of _
#	s = s.encode("ascii","replace")
#	new = s.translate(_TRANS)
#	while True:
#		old = new
#		new = new.replace("__","_")
#		if old == new: break
#	return old

#-------------------------------------------------------------------------------
def header(fout, inp):
	fout.write("// flair converted to openscad\n")
	try:
		title = inp.cards["TITLE"][0]
		fout.write("// Title: %s\n\n"%(title.text()))
	except:
		pass
	fout.write("INF=1000000;\n")
	fout.write("SPH=32;\n")
	fout.write("CYL=32;\n")
	fout.write("\n")

#-------------------------------------------------------------------------------
def exportBodies(fout, inp):
	fout.write("\n// ---- Bodies ----\n")

	bodies = []

	# Loop all cards
	for card in inp.cardlist:
		if not card.isGeo(): continue
		if len(card.tag) != 3: continue
		if card.ignore(): continue
		if card.tag == "END": continue

		name = card.name()
		if name == "": continue
		name = "b_"+name
		bodies.append(name)

		for line in str(card).splitlines():
			fout.write("\n// %s\n"%(line))
		fout.write("module %s()\n{\n"%(name))

		if   card.tag == "RPP":
			fout.write("\ttranslate(%s)\n" % (card.bodyP()))
			fout.write("\t\tcube([%g,%g,%g], center=false);\n"%
				( card.numWhat(2)-card.numWhat(1),
				  card.numWhat(4)-card.numWhat(3),
				  card.numWhat(6)-card.numWhat(5)))

		elif card.tag == "BOX":
			P = card.bodyP()
			X = card.bodyX()
			Y = card.bodyY()
			Z = card.bodyZ()
			lx = X.norm()
			ly = Y.norm()
			lz = Z.norm()
			write_matrix(fout, X,Y,Z,P)
			fout.write("\t\tcube([%g,%g,%g], center=false);\n"%(lx,ly,lz))

		elif card.tag == "SPH":
			fout.write("\ttranslate(%s)\n" % (card.bodyP()))
			fout.write("\t\tsphere(%g,$fn=SPH);\n"%(card.bodyR()))

		elif card.tag == "RCC":
			P = card.bodyP()
			Z = card.bodyZ()
			h = Z.norm()
			write_matrixZ(fout,Z,P)
			fout.write("\tcylinder(h=%g, r=%g, center=false, $fn=CYL);\n" % (h, card.bodyR()))

		elif card.tag == "TRC":
			P = card.bodyP()
			Z = card.bodyZ()
			h = Z.norm()
			write_matrixZ(fout,Z,P)
			fout.write("\tcylinder(h=%g,r1=%g, r2=%g, center=false, $fn=CYL);\n" %
				(h, card.numWhat(7), card.numWhat(8)))

		#elif card.tag == "ELL":
		#	WriteEllipsoid

		#elif card.tag == "WED":
		#	WriteAngleWedge

		#elif card.tag == "ARB"
		#	WritePolyhedron

		elif card.tag == "XYP":
			fout.write("\ttranslate(%s)\n" % (card.bodyP()))
			writePlane0(fout)

		elif card.tag == "XZP":
			fout.write("\ttranslate(%s)\n" % (card.bodyP()))
			fout.write("\trotate(90,[1,0,0])\n")
			writePlane0(fout)

		elif card.tag == "YZP":
			fout.write("\ttranslate(%s)\n" % (card.bodyP()))
			fout.write("\trotate(90,[0,1,0])\n")
			writePlane0(fout)

		elif card.tag == "PLA":
			Z = card.bodyN()
			Z.norm()
			write_matrixZ(fout, Z, card.bodyP())
			writePlane0(fout)

		elif card.tag in ("XCC", "YCC", "ZCC"):
			fout.write("\ttranslate(%s)\n" % (card.bodyP()))
			if   card.tag == "XCC":
			    fout.write("\t\trotate(90,[0,1,0])\n")
			elif card.tag == "YCC":
			    fout.write("\t\trotate(90,[1,0,0])\n")

			fout.write("\t\t\tcylinder(h=INF,r=%g,center=true,$fn=CYL);"%(card.bodyR()))

		elif card.tag == "REC":
			P = card.bodyP()
			X = card.bodyX()
			Y = card.bodyY()
			Z = card.bodyZ()
			lx = X.norm()
			ly = Y.norm()
			lz = Z.norm()
			write_matrix(fout,X,Y,Z,P)
			fout.write("\tscale([%g,%g,1.0])\n"%(lx,ly))
			fout.write("\t\tcylinder(h=%g, r=1, center=false, $fn=CYL);\n"%(lz))

		elif card.tag in ("XEC", "YEC", "ZEC"):
			fout.write("\ttranslate(%s)\n" % (card.bodyP()))
			if   card.tag == "XCC":
			    fout.write("\t\trotate(90,[0,1,0])\n")
			elif card.tag == "YCC":
			    fout.write("\t\trotate(90,[1,0,0])\n")

			fout.write("\t\tscale([%g,%g,1.0])\n"%(card.numWhat(3),card.numWhat(4)))
			fout.write("\t\t\tcylinder(h=INF, r=1, center=true, $fn=CYL);\n")

		else:
			raise Exception("Unknown body %s"%(card.tag))

		fout.write("\n}\n")
	return bodies

#-------------------------------------------------------------------------------
def _exportRegion(fout, region, expr):
	# Find material
	try:
		material = region[Card.MATERIAL]
	except:
		log.error("\nError: region %s is not assigned material\n"%(region.name()))
		return True

	try: density = float(material.what(3))
	except: density = 0.0
	if density <= _DENSITY_LIMIT: return False

	# Write card
	fout.write("\n// Region: %s\tMaterial: %s\n\n"%(region.name(), material.name()))
	for line in str(region).splitlines():
		fout.write("// %s\n"%(line))

	fout.write("module %s()\n{\n"%(region.name()))

	# Check expression
	exp = csg.tokenize(expr)
	try:
		csg.exp2rpn(exp)
		if exp and not csg.check(exp):
			log.error("REGION %s has invalid expression"%(region.name()))
		else:
			stack = []
			for token in exp:
				if token in ("-", "+", "|"): #, "@"):
					if   token == "+": op = "intersection() {"
					elif token == "-": op = "difference() {"
					elif token == "|": op = "union() {"
					b = stack.pop()
					a = stack.pop()

					if isinstance(a,list) and a[0]==op:
						if isinstance(b,str):
							a.insert(len(a)-1, "\t"+b)
						else:
							a[len(a)-1:len(a)-1] = ["\t"+x for x in b]
						lst = a

					elif isinstance(b,list) and b[0]==op and token in ("+","|"):
						if isinstance(a,str):
							b.insert(1, "\t"+a)
						else:
							b[1:1] = ["\t"+x for x in a]
						lst = b

					else:
						lst = [op]
						if isinstance(a,str):
							lst.append("\t"+a)
						else:
							for i in a: lst.append("\t"+i)
						if isinstance(b,str):
							lst.append("\t"+b)
						else:
							for i in b: lst.append("\t"+i)
						lst.append("}")

					stack.append(lst)

				else:
					stack.append("b_%s();"%(token))

	except csg.CSGException:
		log.error("Unbalanced parenthesis")

	lst = stack.pop()
	if isinstance(lst,str):
		fout.write("\t%s\n"%(lst))
	else:
		for i in lst:
			fout.write("\t%s\n"%(i))
	fout.write("}\n")

	return True

#-------------------------------------------------------------------------------
def exportRegions(fout, fluka):
	fout.write("\n\n// ------ Regions ------\n")
	fluka.regionProperties()

	region = None
	expr   = ""
	regions = []
	for card in fluka.cards["REGION"]:
		if card.ignore(): continue
		if card.name()=="&":
			# append to previous expression
			expr += card.expr()
			continue

		if region is None:
			# Remember starting region
			region = card
			expr   = region.expr()
			continue

		if _exportRegion(fout, region, expr):
			regions.append(region.name())

		# Remember starting region
		region = card
		expr   = region.expr()

	# Last region
	if region:
		if _exportRegion(fout, region, expr):
			regions.append(region.name())

	return regions

#-------------------------------------------------------------------------------
def export(fluka, filename):
	fout = open(filename, "w")
	fluka.preprocess()
	header(fout, fluka)
	bodies = exportBodies(fout, fluka)

	for body in bodies:
		fout.write("//%s();\n"%(body))

	regions = exportRegions(fout, fluka)

	fout.write("\n")
	for region in regions:
		fout.write("%s();\n"%(region))
	fout.close()
