# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   8-Jan-2019

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import re
import math

import log
import csg
import bmath

from deck    import Input, Card
from fluka   import Fluka
from moira   import Moira, MoiraCard
from element import Element

from .convert import Convert, register

#===============================================================================
# Conversion class from Fluka to Moira
#===============================================================================
class Fluka2Moira(Convert):
	src = Fluka.MODE
	dst = Moira.MODE
	INVMATRIX = "%s{-1}"	# pattern for inverse matrix

	#-----------------------------------------------------------------------
	def __init__(self, inp, defines=None):
		super().__init__(inp, defines)
		self.input.preprocess(defines)

		self._rotdefi = self.input.getTransformation()
		self.input.bodyProperties(self._rotdefi)
		self.input.regionProperties()#self._rotdefi)
		self._materials = {}		# keep track of defined materials
		self._rotid = {}
		self.output = None

	#-----------------------------------------------------------------------
	def addCard(self, tag, what=None, comment=""):
		card = MoiraCard(tag, what, comment)
		self.add(card)
		return card

	#-----------------------------------------------------------------------
	def material(self, name):
		try:
			return self._materials[name]
		except KeyError:
			try:
				g4mat = Moira.materialFromFluka(name)
				self._materials[name] = g4mat	# remember for next time
				return g4mat
			except KeyError:
				log.error(f"Material {name} was not converted to moira")
		return "G4_Galactic"

	#-----------------------------------------------------------------------
	# Convert materials from fluka to moira
	#-----------------------------------------------------------------------
	def convertMaterials(self):
		# 1st pass register all fluka defined materials
		for mat in self.input["MATERIAL"]:
			if mat.ignore(): continue
			# Check if material is an element existing in G4
#			if mat.what(2)!="" or mat.what(6)!="" or \
#			   not Moira.materialDefaultInFluka(mat.name()):
			self._materials[mat.name()] = mat.name()

		for mat in self.input["MATERIAL"]:
			if mat.ignore(): continue
			mat[Card.USED] = True
			name = mat.name()
			Z    = mat.intWhat(1)
			Am   = mat.intWhat(2)
			rho  = mat.numWhat(3)
			if mat.what(5): self.notImplemented(mat, 5)
			A    = mat.intWhat(6)
			if A != 0:
				self.notImplemented(mat, 6)
#				self.addCard("ISOT", [name, Z, Am, A])
#				self.addCard("MAT", [name, name,
			if Am:
				self.addCard("MATE", [name,Z,Am,rho], mat.comment())
			else:
				# Create mixture by finding the compound
				tag = None
				what = [name, rho, 0 ]	# 3rd element is the # of components
				for card in self.input["COMPOUND"]:
					card[Card.USED] = True
					if card.ignore(): continue
					if card.name() != name: continue
					# normalize fractions if needed
					total = abs(sum([card.numWhat(x) for x in range(1,card.nwhats(),2)]))
					for i in range(1,card.nwhats(),2):
						frac = card.numWhat(i)
						elem = card.what(i+1)
						if frac==0 or not elem: break
						if tag is None:
							if frac>0:
								tag = "MIXT_BY_NATOMS"
							elif elem[0]=="-":
								tag = "MIXT_BY_VOLUME"
							else:
								tag = "MIXT"
						if tag != "MIXT_BY_NATOMS":
							frac = bmath.format(abs(frac)/total)
						if elem[0]=="-": elem = elem[1:]
						what.append(self.material(elem))
						what.append(frac)
				if tag is not None:
					what[2] = (len(what)-3)//2
					self.addCard(tag, what, mat.comment())
				elif Z != 0:
					# Find A from database and create material
					element = Element.byZ(Z)
					self.addCard("MATE",
							[name,Z,element.Amass(),rho],
							mat.comment())

	#-----------------------------------------------------------------------
	def convertTransformations(self):
		# find unique names
		self._rotid = {}
		for name, rot in self._rotdefi.items():
			if not isinstance(rot,int): continue
			if rot in self._rotid: continue
			self._rotid[rot] = name
			rotm = self._rotdefi[rot]
#			rx,ry,rz = [bmath.format(math.degrees(x))
#					for x in rotm.getEulerRotation()]
#			self.addCard("ROTM", [name, rx, ry, rz])
			what = [name]
			for i in range(3):
#				what.extend(rotm[i][:3])
				what.extend([bmath.format(x) for x in rotm[i][:3]])
			self.addCard("ROTM", what)
#			print()
#			print(name)
#			print("M=",rotm)
#			print("Rxyz=",rx,ry,rz, "T=",rotm[0][3],rotm[1][3],rotm[2][3])

			# and the inverse matrix
			roti = rotm.clone()
			roti.inv()
#			rx,ry,rz = [bmath.format(math.degrees(x),16)
#					for x in roti.getEulerRotation()]
#			self.addCard("ROTM", [Fluka2Moira.INVMATRIX%(name), rx, ry, rz])
			what = [Fluka2Moira.INVMATRIX%(name)]
			for i in range(3):
#				what.extend(roti[i][:3])
				what.extend([bmath.format(x) for x in roti[i][:3]])
			self.addCard("ROTM", what)
#			print("I=",roti)
#			print("Rxyz=",rx,ry,rz, "T=",roti[0][3],roti[1][3],roti[2][3])
#			print()
#			print()
#			print()
#			print()

	#-----------------------------------------------------------------------
#	def findBody(self, name):
#		for t in CardInfo.bodies[self.input.MODE]:
#			for x in self.input[t]:
#				if x.notIgnore() and x.name()==name:
#					return x

	#-----------------------------------------------------------------------
	def findBlackBody(self):
		# --- Find world and void ---
		# Find outer blackbody
		# FIXME very very primitive
		# it will search for:
		#  1. regions with BLCKHOLE as material
		#  2. name: BLKBODY
		#  3. in expression to find +BODY -[...]
		#     where BODY is referred only in this expression
		#  4. typically the first one or the last one
		# it should be the biggest region with
		# BLACKHOLE material and typically a simple expression +BLK -VOID
		# where both BLK and VOID are simple objects RPP or BOX.

		# build dictionary with biggest dimension
		bodyMax = {}
		bodies = self.input.cardsCache(Card.BODY)
		for name,body in bodies.items():
			m = max([abs(body.numWhat(x))
						for x in range(1,body.nwhats())])
			bodyMax[name] = m, body

		# For all regions with BLCKHOLE as material
		# find the biggest body
		regionMax = 0
		blkregion = None
		for region in self.input["REGION"]:
			if region.ignore(): continue
			material = region[Card.MATERIAL]
			if material is not None and \
			   material.name() == "BLCKHOLE":
				for token in csg.tokenize(region.expr()):
					if token in ("(","+","-","|",")"): continue
					try:
						r, body = bodyMax[token]
						if r>regionMax:
							regionMax = r
							blkregion = region
					except KeyError:
						pass

		if blkregion is None:
			raise Exception("Black body search didn't work out!")

		# Possible patterns of blkregion
		# 1. +blkbody -world
		expr = "".join(blkregion.expr().split())
		m = re.match(r"^\+(\w+)-(\w+)$", expr)
		if m:
			world = bodies[m.group(2)]
		else:
			# 2. -blkbody
			m = re.match(r"^-(\w+)$", blkregion.expr())
			if m:
				world = bodies[m.group(1)]
			else:
				# 3. +blkbody -body1 -body2 ....
				m = re.match(r"^\+(\w+)\b", expr)
				if m:
					world = bodies[m.group(1)]
				else:
					raise Exception("Black body search didn't work out!")

		if world.tag not in ("RPP","SPH"):
			raise Exception("Black body seems it is not RPP or SPH")
		log.say("Finding blackbody:")
		log.say("\tBlkregion=", blkregion)
		log.say("\tWorld=", world)
		return blkregion, world

	#-----------------------------------------------------------------------
	def convertBodies(self, world):
		matrix = None
		numMat=0
		for card in self.input.cardlist:
			if card.ignore():    continue
			if not card.isGeo(): continue
#			if card is world:
#				card[Card.USED] = True
#			elif card.type() == Card.BODY:
			if card.type() == Card.BODY:
				if card[Card.MATRIX] is not matrix:
					matrix = card[Card.MATRIX]
					if matrix is not None:
						#FIXME Find a better idea..
						what = [f"MATRIX{numMat}"]
						numMat += 1
						for row in matrix[:-1]: # skip last 0 0 0 1 row
							what.extend(row)
						self.add(MoiraCard("MATRIX", what))
					else:
						self.add(MoiraCard("MATRIX"))
				self.add(MoiraCard(card.tag, card.whats()[:], card.comment()))
				card[Card.USED] = True
			elif card.tag[0] == "#":
				self.add(card.clone())
				card[Card.USED] = True
#			elif card.tag[0] == "$":
#				card[Card.USED] = True

	#-----------------------------------------------------------------------
	# Convert them to volumes
	#-----------------------------------------------------------------------
	def volumeCard(self, body, parent=None, mat="G4_Galactic", comment=""):
		name = body.name()
		if body.tag == "SPH":
			# Comm
			card = MoiraCard("VOLU.ORB", [name, mat, body.numWhat(4)], comment)
			x = body.numWhat(1)
			y = body.numWhat(2)
			z = body.numWhat(3)
		elif body.tag == "RPP":
			dx = (body.numWhat(2)-body.numWhat(1))/2
			dy = (body.numWhat(4)-body.numWhat(3))/2
			dz = (body.numWhat(4)-body.numWhat(5))/2
			card = MoiraCard("VOLU.BOX", [name, mat, dx, dy, dz], comment)
			x = bmath.format((body.numWhat(2)+body.numWhat(1))/2)
			y = bmath.format((body.numWhat(4)+body.numWhat(3))/2)
			z = bmath.format((body.numWhat(4)+body.numWhat(5))/2)
		else:
			raise Exception(name+" with type "+body.tag+" is not supported as volume")
		self.add(card)
		if parent:
			card = MoiraCard("PLACE", [name, 1, parent, "", x, y, z], comment)
			self.add(card)

	#-----------------------------------------------------------------------
	# Regions & Volumes
	#-----------------------------------------------------------------------
	def convertRegions(self, blkregion, world):
		# Write the basic volumes
		self.volumeCard(world, comment=world.comment())
		#self.writeline(":VIS %s OFF"%(world.name()))

		#self.volumeCard(void, world.name(), comment=void.comment())
		#self.writeline(":VIS %s OFF"%(void.name()))

		magRegions = []
		for card in self.input["ASSIGNMAT"]:
			if card.intWhat(5) in (1,4):
				fromreg = card.what(2)
				toreg   = card.what(3)
				fromnum = [int(s) for s in re.findall(r'-?\d+\.?\d*', fromreg)]
				tonum   = [int(s) for s in re.findall(r'-?\d+\.?\d*', toreg)]
				if fromnum and tonum:
					base = fromreg.replace(str(fromnum[0]),'')
					if fromnum:
						for i in range(fromnum[0], tonum[0]+1):
							magRegions.append(f"{base}{i}")
				else:
					magRegions.append(fromreg)
					if toreg:
						magRegions.append(toreg)
		card = None
		blkregion[Card.USED] = True
		try:
			blkregion[Card.ASSIGNMAT][Card.USED] = True
		except (KeyError, TypeError):
			pass
		for region in self.input["REGION"]:
			if region.ignore() or region==blkregion: continue
			# FIXME
			material = region[Card.MATERIAL]
			try:
				region[Card.ASSIGNMAT][Card.USED] = True
			except (KeyError, TypeError):
				pass
			expr = "\n".join([x.strip() for x in region.expr().splitlines()])
			if region.name()=="&":
				# continuation of previous card
				card.setComment(card.comment()+"\n"+region.comment())
				card.setExpr(card.expr()+"\n"+expr)
				region[Card.USED] = True
			else:
				lattice = region[Card.LATTICE]
				if lattice:
					g4mat = ""
				elif material:
					g4mat = self.material(material.name())
				else:
					g4mat = "G4_Galactic"
				field = int(region.name() in magRegions)
				card = MoiraCard("CELL",[region.name(),
							world.name(),
							g4mat,
							field],
						region.comment())
				card.setExpr(expr)
				self.add(card)
				region[Card.USED] = True
				if lattice:
					name = lattice.name()
					inv = True
					if name[0] == "-":
						inv = not inv
						name = name[1:]
					rotm = self.input.getTransformation(name, self._rotdefi, inv)
					# try to get the real name used based on the id
					rid  = self._rotdefi[name]
					name = self._rotid[rid]
					if inv: name = Fluka2Moira.INVMATRIX%(name)
					dx = rotm[0][3]
					dy = rotm[1][3]
					dz = rotm[2][3]
					self.addCard("LATTICE",
							[region.name(), name,
							dx, dy, dz],
							lattice.comment())

	#-----------------------------------------------------------------------
	def convertAllOtherCards(self):
		# --- Process all other cards ---
		iterator = iter(self.input.cardlist)
		for card in iterator:
			if card.ignore(): continue
			if card.tag in Input._PREPRO_BLOCK: continue

			if isinstance(card, MoiraCard) or card.ignore():
				self.add(card.clone())

			elif card.tag == "BEAM":
				particle = card.name()
				if particle in ("ISOTOPE","HEAVYION"):
					self.notImplemented(card,0)
					continue
				self.add(MoiraCard("/run/initialize"))
				self.add(MoiraCard("/primary/select",["gps"],
						"Primary particle generator"))
				self.add(MoiraCard("/gps/particle",
						[Moira.particleFromFluka(particle)],
						card.comment()))

				# We should keep the formulas if any
				energy = card.numWhat(1)	# or momentum
				if energy>0:
					self.notImplemented(card,"conversion from momentum")
				self.add(MoiraCard("/gps/energy",[-energy,"GeV"]))

				# FIXME particle distribution etc....

			elif card.tag == "BEAMPOS":
				x = card.numWhat(1)
				y = card.numWhat(2)
				z = card.numWhat(3)
				self.add(MoiraCard("/gps/pos/centre",[x,y,z,"cm"],card.comment()))
				u = card.numWhat(4)
				v = card.numWhat(5)
				w = math.sqrt(1-u*u-v*v)
				if card.sdum()=="":
					pass
				elif card.sdum()=="NEGATIVE":
					w = -w
				else:
					self.notImplemented(card,0)
					continue
				self.add(MoiraCard("/gps/direction",[u,v,w]))

			#------------------------------------------------------
			# Biasing
			#------------------------------------------------------
			elif card.tag == "BIASING":
				part  = card.what(1)
				imp   = card.what(3)
				reg   = card.what(4)
				if part == "0":	part = "all"
				if part == "1":	part = "hadrons"
				if part == "2":	part = "em"
				if part == "3":	part = "neutrons"
				self.add(MoiraCard("/bias/region",[reg,part,imp],card.comment()))

			#------------------------------------------------------
			# Magnetic Fields
			# FIXME this specific to our BMI interpretation
			#       we should make somehow a plugin
			#------------------------------------------------------
			elif card.tag == "USRGCALL":
				reg   = card.what(2)
				step  = card.what(4)
				grad  = card.what(5)
				offX  = card.what(7)
				offY  = card.what(8)
				offZ  = card.what(9)
				alpha = card.what(10)
				beta  = card.what(11)
				gamma = card.what(12)
				if card.sdum() == "&": continue # FIXME
				typeB  = "dipole"
				coreB  = "dipole"
				simX   = "false"
				simY   = "false"
				simZ   = "false"
				simR   = "false"
				Bx     = 0.0
				By     = 0.0
				Bz     = 0.0
				r      = 0.0
				xmin   = 0
				xmax   = 0
				xbins  = 0
				ymin   = 0
				ymax   = 0
				ybins  = 0
				mapInp = f"{card.sdum()}.dat"
				try:
					fileMap = open(mapInp,'r')
				except IOError:
					continue
				for line in fileMap:
					lineArr = line.split()		#FIXME too ugly
					if lineArr[0][0] == "#":
						continue
					if lineArr[0] == "TYPE":
						typeB = lineArr[1]
						print(lineArr)
					elif lineArr[0] == "SYMMETRY":
						simX = "true" if 'X' in lineArr[1] else "false"
						simY = "true" if 'Y' in lineArr[1] else "false"
						simZ = "true" if 'Z' in lineArr[1] else "false"
						print(lineArr)
					elif lineArr[0] == "QORIGIN":
						Bx = lineArr[1]
						By = lineArr[2]
						Bz = lineArr[3]
						print(lineArr)
					elif lineArr[0] == "QRADIUS":
						r  = lineArr[1]
						print(lineArr)
					elif lineArr[0] == "XGRID":
						xmin  = lineArr[1]
						xmax  = lineArr[2]
						xbins = lineArr[3]
					elif lineArr[0] == "YGRID":
						ymin  = lineArr[1]
						ymax  = lineArr[2]
						ybins = lineArr[3]

					#next(iterator)
					#print(card.sdum())
					#if (card.sdum() == "&"):
						#print(card.sdum())
						#offX = card.what(1)
						#offY = card.what(2)
						#offZ = card.what(3)
						#alpha = card.what(4)
						#beta  = card.what(5)
						#gamma = card.what(6)
					if typeB == "INTER2D":
						# FIXME introduce multiple cards with less parameters
						self.add(MoiraCard("/field/magnetic/interpolated2D/region",
								[reg, grad, "tesla", coreB,
								"DormandPrince745", step, offX,
								offY, offZ, Bx, By, Bz,
								alpha, beta, gamma, mapInp,
								r, xmin, xmax, xbins,
								ymin, ymax, ybins, simX,
								simY, simZ, simR],card.comment()))
					elif typeB == "RZ":
						# FIXME introduce multiple cards with less parameters
						self.add(MoiraCard("/field/magnetic/interpolated2D/region",
								[reg, grad, "tesla", coreB,
								"DormandPrince745", step, offX,
								offY, offZ, Bx, By, Bz,
								alpha, beta, gamma, mapInp,
								r, xmin, xmax, xbins,
								ymin, ymax, ybins, simX,
								simY, simZ, "true"],card.comment()))
					elif typeB == "QUADINT":
						# FIXME introduce multiple cards with less parameters
						self.add(MoiraCard("/field/magnetic/interpolated2D/region",
								[reg, grad, "tesla", "quadrupole",
								"DormandPrince745", step, offX,
								offY, offZ, Bx, By, Bz,
								alpha, beta, gamma, mapInp,
								r, xmin, xmax, xbins,
								ymin, ymax, ybins, simX,
								simY, simZ, simR],card.comment()))
					elif typeB == "INTER3D":
						# FIXME introduce multiple cards with less parameters
						self.add(MoiraCard("/field/magnetic/interpolated3D/region",
								[reg, grad, "tesla", coreB,
								"DormandPrince745", step, offX,
								offY, offZ, Bx, By, Bz,
								alpha, beta, gamma, mapInp,
								r, xmin, xmax, xbins,
								ymin, ymax, ybins, zmin,
								zmax, zbins, simX, simY,
								simZ],card.comment()))
					else:
						if   card.sdum() == "QUAD":  coreB = "quadrupole"
						elif card.sdum() == "SEXTU": coreB = "sextupole"
						elif card.sdum() == "OCTU":  coreB = "octupole"
						elif card.sdum() == "DECA":  coreB = "decapole"
						self.add(MoiraCard("/field/magnetic/perfect/region",
								[reg, grad, "tesla", coreB,
								"DormandPrince745", step, offX,
								offY, offZ, Bx, By, Bz,
								alpha, beta, gamma, simX,
								simY, simZ],card.comment()))
					if lineArr[0] == "DATA":
						break
				fileMap.close()

			elif card.tag == "USRBIN":
				# FIXME
				# - combine same dimension/type meshes under common mesh
				name  = card.name()
				utype = card.intWhat(1)
				#part  = card.what(2)
				#unit  = abs(card.intWhat(3))
				xmax  = card.numWhat(4)
				ymax  = card.numWhat(5)
				zmax  = card.numWhat(6)
				xmin  = card.numWhat(7)
				ymin  = card.numWhat(8)
				zmin  = card.numWhat(9)
				nx    = max(card.intWhat(10),1)
				ny    = max(card.intWhat(11),1)
				nz    = max(card.intWhat(12),1)

				if utype in (0,10):
					new = MoiraCard("MESH.BOX",[name],card.comment())
					x  = (xmin+xmax)/2
					y  = (ymin+ymax)/2
					z  = (zmin+zmax)/2
					new["boxSize.dx"] = (xmax-xmin)/2
					new["boxSize.dy"] = (ymax-ymin)/2
					new["boxSize.dz"] = (zmax-zmin)/2
					new["boxSize.Unit"] = "cm"
					new["nBin.Nx"] = nx
					new["nBin.Ny"] = ny
					new["nBin.Nz"] = nz
				elif utype in (1,11):
					new = MoiraCard("MESH.CYLINDER",[name],card.comment())
					# FIXME no rmin
					#r  = xmax
					x  = ymin
					y  = ymax
					z  = (zmin+zmax)/2
					new["cylinderSize.R"]  = xmax
					new["cylinderSize.dz"] = (zmax-zmin)/2
					new["cylinderSize.Unit"] = "cm"
					new["nBin.Nr"]   = nx
					new["nBin.Nz"]   = nz
					new["nBin.Nphi"] = ny
				else:
					self.notImplemented(card,1)
					clone = card.clone()
					clone.setEnable(False)
					self.add(clone)
					continue

				new["xyz.X"] = x
				new["xyz.Y"] = y
				new["xyz.Z"] = z
				new["xyz.Unit"] = "cm"
				self.add(new)

# FIXME create groups of particles as particle-lists
#
#/score/quantity/energyDeposit enedep MeV
#/score/quantity/cellFlux neutron percm2
#/score/filter/particle . neutron
#/score/quantity/cellFlux electron percm2
#/score/filter/particle . e+ e-
				self.add(MoiraCard("/score/close"))

			elif card.tag == "RANDOMIZ" and card.what(2):
				seed = ("%s %s %s"%(card.what(2), card.what(2), card.what(2))).strip()
				self.add(MoiraCard("/random/setSeeds",[seed]))

			elif card.tag == "START":
				self.add(MoiraCard("/run/beamOn",[card.intWhat(1)]))

			elif card[Card.USED] or \
			     card.tag in ("GEOBEGIN", "GEOEND", "STOP", "END"):
				pass

			else:
				clone = card.clone()
				clone.setEnable(False)
				self.add(clone)

	#-----------------------------------------------------------------------
	# FIXME some automatic unit conversion system!
	#-----------------------------------------------------------------------
	def convert(self):
		self.output = Moira()
		try:
			card = self.input["TITLE"][0]
			card[Card.USED] = True
			self.addCard("/moira/title",[card.text()],card.comment())
		except IndexError:
			pass
		self.addCard("/physics_lists/select")

		self.convertMaterials()
		self.convertTransformations()
		blkregion, world = self.findBlackBody()
		self.convertBodies(world)
		self.convertRegions(blkregion, world)
		self.convertAllOtherCards()
		return self.output

#-------------------------------------------------------------------------------
# Execute only once and register the conversion program
#-------------------------------------------------------------------------------
register(Fluka.MODE, Moira.MODE, Fluka2Moira)
