# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   24-Oct-2018

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

__all__ = [	"Fluka",
		"FlukaCard",
		"FlukaCardInfo",
		"FlukaConfig",
		"FlukaLayout",
		"FlukaVoxel",
		"Particle",
		"Voxel" ]

from itertools import count

from log import say

from .card     import FlukaCard, FlukaCardInfo
from .input    import Fluka, LowNeutMaterial
from .particle import Particle
from .voxel    import FlukaVoxel
from .layout   import FlukaLayout
from .config   import FlukaConfig

DATABASE = "db/fluka.ini"

#-------------------------------------------------------------------------------
# Initialize classes:
#	FlukaCardInfo
#	FlukaParticle
#-------------------------------------------------------------------------------
def init(filename=None):
	import re
	import configparser
	from deck import Card, CardInfo, dbpath, readCardini

	Card.register_class(Fluka.MODE, FlukaCard)

	# Load fluka database
	if filename is None:
		filename = dbpath(DATABASE)

	# read card database and prepare CardInfo classes
	cardini = configparser.RawConfigParser()
	if not cardini.read(filename): return False

	# parse cards
	readCardini(cardini, Fluka.MODE, FlukaCardInfo)

	# Geometry transformation tags
	FlukaCardInfo.TRANSFORM_TAGS = [ x.tag for x in CardInfo.cards(Fluka.MODE)
					if x.tag[0]=="$" ]
	FlukaCardInfo.TRANSFORM_TAGS.sort()

	# Particles
	section = "@particles"
	for pid, name in cardini.items(section):
		if pid[:3] == "pid":
			tag, idx = pid.split(".")
			mass    = cardini.getfloat(section,"mass."+idx)
			comment = cardini.get(section,"comment."+idx)
			try:
				pdg = int(cardini.get(section,"pdg."+idx))
			except:
				pdg = 0
			Particle.add(int(idx), name, mass, comment, pdg)
	Particle.makeLists()

	# Read default materials and create fake MATERIAL card's
	section = "@materials"
	for i in count(1):
		n = "name.%d" % (i)
		try: name = cardini.get(section, n)
		except: break
		desc  = cardini.get(     section, "desc.%d"  % (i))
		Amass = cardini.getfloat(section, "Amass.%d" % (i))
		Z     = cardini.getfloat(section, "Z.%d"     % (i))
		rho   = cardini.getfloat(section, "rho.%d"   % (i))
		mat   = FlukaCard("MATERIAL",[name, Z, Amass, rho, i],desc)
		mat["@n"] = i
		Fluka.defaultMaterials[name] = mat

	# Read special ICRU materials
	for i in count(-1,-1):
		n = "name.%d" % (i)
		try: name = cardini.get(section, n)
		except: break

		desc  = cardini.get(     section, "desc.%d"  % (i))
		Amass = cardini.getfloat(section, "Amass.%d" % (i))
		Z     = cardini.getfloat(section, "Z.%d"     % (i))
		rho   = cardini.getfloat(section, "rho.%d"   % (i))
		mat   = FlukaCard("MATERIAL",[name, Z, Amass, rho],desc)
		mat["@n"] = i
		Fluka.icruMaterials[name] = mat

	# Low neutrons energy groups
	section = "@n-groups"
	# Read number of groups
	try:
		groups = cardini.get(section,"groups")
		Fluka.NGROUPS = list(map(int, groups.split()))
	except:
		say("ERROR! No neutron energy groups defined")

	pat = re.compile(r"^(\S*) *(.*)$")
	for g in Fluka.NGROUPS:
		groupsList  = []
		groupsListS = []
		for i in range(1,g+2):
			n = "ene.%d.%d" % (g,i)
			try: energy = cardini.get(section, n)
			except: break
			m = pat.match(energy)
			groupsList.append(float(m.group(1)))
			groupsListS.append(m.group(2))
		Fluka.neutronGroups[g]  = groupsList
		Fluka.neutronGroupsS[g] = groupsListS

	# Low neutrons energy groups
	for g in Fluka.NGROUPS:
		section   = "@low-neut"
		materials = []
		for i in count(1):
			n = "elem.%d.%d" % (g,i)
			try: elem = cardini.get(section, n)
			except: break
			mat = LowNeutMaterial()
			mat.elem = elem
			mat.desc = cardini.get(section, "desc.%d.%d"  % (g,i))
			mat.name = cardini.get(section, "name.%d.%d" % (g,i))
			mat.temp = cardini.get(section, "temp.%d.%d" % (g,i))
			mat.db   = cardini.get(section, "db.%d.%d"   % (g,i))
			ids      = cardini.get(section, "ids.%d.%d"  % (g,i)).split()
			mat.id1  = int(ids[0])
			mat.id2  = int(ids[1])
			mat.id3  = int(ids[2])
			mat.g    = cardini.get(section, "g.%d.%d" % (g,i))
			materials.append(mat)
		Fluka.lowMaterials[g] = materials

	# Read user routines
	section = "@user"
	for routine, desc in cardini.items(section):
		FlukaConfig.user[routine] = desc
	del cardini
	return True

#===============================================================================
# Load only once
#init()
